import { generateComponentId, decodeComponentId, factoryWithDefaultOptions } from '@previewjs/component-analyzer-api';
import { EMPTY_OBJECT_TYPE, UNKNOWN_TYPE, dereferenceType, objectType, maybeOptionalType, stripUnusedTypes, helpers, NODE_TYPE, functionType, arrayType, createTypeAnalyzer } from '@previewjs/type-analyzer';
import { createStackedReader, createFileSystemReader } from '@previewjs/vfs';
import path from 'path';
import ts from 'typescript';
import url from 'url';
import { parseSerializableValue } from '@previewjs/serializable-values';
import { extractStoriesInfo, extractArgs, extractCsf3Stories, resolveComponentId } from '@previewjs/storybook-helpers';

function detectPropTypes(sourceFile, name) {
  for (const statement of sourceFile.statements) {
    if (ts.isExpressionStatement(statement) && ts.isBinaryExpression(statement.expression) && statement.expression.operatorToken.kind === ts.SyntaxKind.EqualsToken && ts.isPropertyAccessExpression(statement.expression.left) && ts.isIdentifier(statement.expression.left.expression)) {
      const componentName = statement.expression.left.expression.text;
      if (componentName !== name) {
        continue;
      }
      if (statement.expression.left.name.text === "propTypes") {
        return statement.expression.right;
      }
    } else if (ts.isClassDeclaration(statement) && statement.name) {
      const componentName = statement.name.text;
      if (componentName !== name) {
        continue;
      }
      for (const classMember of statement.members) {
        if (ts.isPropertyDeclaration(classMember) && classMember.initializer && classMember.name.getText() === "propTypes") {
          return classMember.initializer;
        }
      }
    }
  }
  return null;
}

function analyzeReactComponent(logger, typeResolver, absoluteFilePath, componentName, signature) {
  const sourceFile = typeResolver.sourceFile(absoluteFilePath);
  let propTypes = null;
  if (sourceFile) {
    propTypes = detectPropTypes(sourceFile, componentName);
  }
  const resolved = computePropsType(logger, typeResolver, signature, propTypes);
  return {
    props: resolved.type,
    types: { ...resolved.collected }
  };
}
function computePropsType(logger, typeResolver, signature, propTypes) {
  if (propTypes) {
    return computePropsTypeFromPropTypes(typeResolver, propTypes);
  }
  return computePropsTypeFromSignature(logger, typeResolver, signature);
}
function computePropsTypeFromSignature(logger, typeResolver, signature) {
  const firstParam = signature.getParameters()[0];
  if (!firstParam) {
    return {
      type: EMPTY_OBJECT_TYPE,
      collected: {}
    };
  }
  if (!firstParam.valueDeclaration) {
    return {
      type: UNKNOWN_TYPE,
      collected: {}
    };
  }
  const type = typeResolver.checker.getTypeOfSymbolAtLocation(
    firstParam,
    firstParam.valueDeclaration
  );
  try {
    let { type: props, collected } = typeResolver.resolveType(type);
    [props] = dereferenceType(props, collected, []);
    stripUnusedProps:
      if (props.kind === "object" && ts.isParameter(firstParam.valueDeclaration)) {
        if (ts.isObjectBindingPattern(firstParam.valueDeclaration.name)) {
          const bindingPattern = firstParam.valueDeclaration.name;
          const usedProps = /* @__PURE__ */ new Set();
          const propsWithDefault = /* @__PURE__ */ new Set();
          for (const element of bindingPattern.elements) {
            if (element.dotDotDotToken) {
              break stripUnusedProps;
            }
            const elementName = element.propertyName || element.name;
            if (!ts.isIdentifier(elementName)) {
              break stripUnusedProps;
            }
            usedProps.add(elementName.text);
            if (element.initializer) {
              propsWithDefault.add(elementName.text);
            }
          }
          props = objectType(
            Object.fromEntries(
              Object.entries(props.fields).filter(([key]) => usedProps.has(key)).map(([key, type2]) => [
                key,
                maybeOptionalType(type2, propsWithDefault.has(key))
              ])
            )
          );
        }
      }
    return { type: props, collected };
  } catch (e) {
    logger.warn(
      `Unable to resolve props type for ${typeResolver.checker.typeToString(
        type
      )}`,
      e
    );
  }
  return {
    type: UNKNOWN_TYPE,
    collected: {}
  };
}
function computePropsTypeFromPropTypes(typeResolver, propTypes) {
  const type = typeResolver.checker.getTypeAtLocation(propTypes);
  const fields = {};
  let collected = {};
  for (const property of type.getProperties()) {
    fields[property.name] = (() => {
      if (!property.valueDeclaration) {
        return UNKNOWN_TYPE;
      }
      const propertyType = typeResolver.checker.getTypeAtLocation(
        property.valueDeclaration
      );
      const typeArguments = typeResolver.resolveTypeArguments(propertyType);
      const fieldType = typeArguments.types[0];
      if (fieldType) {
        collected = {
          ...collected,
          ...typeArguments.collected
        };
        return maybeOptionalType(
          fieldType,
          propertyType.symbol.name === "Requireable"
        );
      }
      return UNKNOWN_TYPE;
    })();
  }
  const resolved = objectType(fields);
  return {
    type: resolved,
    collected: stripUnusedTypes(collected, [resolved])
  };
}

function extractReactComponents(logger, resolver, rootDir, absoluteFilePath) {
  const sourceFile = resolver.sourceFile(absoluteFilePath);
  if (!sourceFile) {
    return [];
  }
  const functions = [];
  for (const statement of sourceFile.statements) {
    if (ts.isExportAssignment(statement)) {
      if (ts.isIdentifier(statement.expression)) {
        continue;
      }
      functions.push(["default", statement, statement.expression]);
    } else if (ts.isVariableStatement(statement)) {
      for (const declaration of statement.declarationList.declarations) {
        if (!ts.isIdentifier(declaration.name) || !declaration.initializer) {
          continue;
        }
        functions.push([
          declaration.name.text,
          statement,
          declaration.initializer
        ]);
      }
    } else if (ts.isFunctionDeclaration(statement)) {
      const isDefaultExport = !!statement.modifiers?.find(
        (m) => m.kind === ts.SyntaxKind.ExportKeyword
      ) && !!statement.modifiers?.find(
        (m) => m.kind === ts.SyntaxKind.DefaultKeyword
      );
      const name = statement.name?.text;
      if (isDefaultExport || name) {
        functions.push([name || "default", statement, statement]);
      }
    } else if (ts.isClassDeclaration(statement) && statement.name) {
      functions.push([statement.name.text, statement, statement]);
    }
  }
  const storiesInfo = extractStoriesInfo(sourceFile);
  const components = [];
  const args = extractArgs(sourceFile);
  const nameToExportedName = helpers.extractExportedNames(sourceFile);
  function extractComponent(baseComponent, node, name) {
    if (name === "default" && storiesInfo) {
      return null;
    }
    const storyArgs = args[name];
    const isExported = name === "default" || !!nameToExportedName[name];
    const signature = extractComponentSignature(resolver.checker, node);
    if (storiesInfo && isExported && (storyArgs || signature?.parameters.length === 0)) {
      const associatedComponent = extractStoryAssociatedComponent(
        logger,
        resolver,
        rootDir,
        storiesInfo.component
      );
      return {
        ...baseComponent,
        kind: "story",
        args: storyArgs ? {
          start: storyArgs.getStart(),
          end: storyArgs.getEnd(),
          value: parseSerializableValue(storyArgs)
        } : null,
        associatedComponent
      };
    }
    if (signature) {
      return {
        ...baseComponent,
        kind: "component",
        exported: isExported,
        extractProps: async () => analyzeReactComponent(
          logger,
          resolver,
          absoluteFilePath,
          name,
          signature
        )
      };
    }
    return null;
  }
  for (const [name, statement, node] of functions) {
    const component = extractComponent(
      {
        componentId: generateComponentId({
          filePath: path.relative(rootDir, absoluteFilePath),
          name
        }),
        offsets: [statement.getStart(), statement.getEnd()]
      },
      node,
      name
    );
    if (component) {
      components.push(component);
    }
  }
  return [
    ...components,
    ...extractCsf3Stories(
      rootDir,
      resolver,
      sourceFile,
      async (componentId) => {
        const { filePath } = decodeComponentId(componentId);
        const component = extractReactComponents(
          logger,
          resolver,
          rootDir,
          path.join(rootDir, filePath)
        ).find((c) => c.componentId === componentId);
        if (component?.kind !== "component") {
          return {
            props: UNKNOWN_TYPE,
            types: {}
          };
        }
        return component.extractProps();
      }
    )
  ];
}
function extractStoryAssociatedComponent(logger, resolver, rootDir, component) {
  const resolvedStoriesComponentId = resolveComponentId(
    rootDir,
    resolver.checker,
    component
  );
  return component && resolvedStoriesComponentId ? {
    componentId: resolvedStoriesComponentId,
    extractProps: async () => {
      const signature = extractComponentSignature(
        resolver.checker,
        component
      );
      if (!signature) {
        return {
          props: UNKNOWN_TYPE,
          types: {}
        };
      }
      const { filePath, name } = decodeComponentId(
        resolvedStoriesComponentId
      );
      return analyzeReactComponent(
        logger,
        resolver,
        path.join(rootDir, filePath),
        name,
        signature
      );
    }
  } : null;
}
function extractComponentSignature(checker, node) {
  let type = checker.getTypeAtLocation(node);
  if (type.flags === ts.TypeFlags.Any && ts.isCallExpression(node) && ts.isPropertyAccessExpression(node.expression) && ts.isIdentifier(node.expression.name) && node.expression.name.text === "bind") {
    const symbol = checker.getSymbolAtLocation(node.expression.expression);
    if (symbol?.valueDeclaration && ts.isVariableDeclaration(symbol.valueDeclaration) && symbol.valueDeclaration.initializer) {
      type = checker.getTypeAtLocation(symbol.valueDeclaration.initializer);
    }
  }
  for (const callSignature of type.getCallSignatures()) {
    if (isValidComponentReturnType(callSignature.getReturnType())) {
      return callSignature;
    }
  }
  if (type.symbol) {
    const classType = checker.getTypeOfSymbolAtLocation(type.symbol, node);
    for (const constructSignature of classType.getConstructSignatures()) {
      const returnType = constructSignature.getReturnType();
      if (returnType.getProperty("render")) {
        return constructSignature;
      }
    }
  }
  return null;
}
function isValidComponentReturnType(type) {
  if (isJsxElement(type)) {
    return true;
  }
  return false;
}
const jsxElementTypes = /* @__PURE__ */ new Set(["Element", "ReactElement"]);
function isJsxElement(type) {
  if (type.isUnion()) {
    for (const subtype of type.types) {
      if (isJsxElement(subtype)) {
        return true;
      }
    }
  }
  return jsxElementTypes.has(type.symbol?.getEscapedName().toString());
}

const REACT_SPECIAL_TYPES = {
  Component: NODE_TYPE,
  ComponentClass: functionType(NODE_TYPE),
  FunctionComponent: functionType(NODE_TYPE),
  ReactElement: NODE_TYPE,
  ReactNode: NODE_TYPE,
  ReactNodeArray: arrayType(NODE_TYPE),
  ReactPortal: NODE_TYPE
};

const createComponentAnalyzer = factoryWithDefaultOptions(
  ({ rootDir, reader, logger }) => {
    const __dirname = url.fileURLToPath(new URL(".", import.meta.url));
    const typesDirPath = path.join(__dirname, "..", "types");
    const typeAnalyzer = createTypeAnalyzer({
      reader: createStackedReader([
        reader,
        createFileSystemReader({
          mapping: {
            from: typesDirPath,
            to: path.join(rootDir, "node_modules", "@types")
          },
          watch: false
        })
      ]),
      rootDir,
      specialTypes: REACT_SPECIAL_TYPES,
      tsCompilerOptions: {
        jsx: ts.JsxEmit.ReactJSX,
        jsxImportSource: "react"
      },
      warn: logger.warn.bind(logger)
    });
    return {
      typeAnalyzer,
      detectComponents: async (filePaths) => {
        const absoluteFilePaths = filePaths.map(
          (f) => path.isAbsolute(f) ? f : path.join(rootDir, f)
        );
        const resolver = typeAnalyzer.analyze(absoluteFilePaths);
        const components = [];
        for (const absoluteFilePath of absoluteFilePaths) {
          components.push(
            ...extractReactComponents(
              logger,
              resolver,
              rootDir,
              absoluteFilePath
            )
          );
          await 0;
        }
        return components;
      },
      dispose: () => {
        typeAnalyzer.dispose();
      }
    };
  }
);

export { createComponentAnalyzer };
